"""Return a config object from a json file for the application."""

import json

from .config import Config
import psiconfig.text as text


class JsonConfig(Config):
    """
        A class to handle config files in json format
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _read_config(self) -> dict[str, object]:
        """Open the config file and return the contents as a dict."""
        self.status = self.OK
        try:
            with open(self.path, 'r') as f_config:
                try:
                    config = json.load(f_config)
                    return self.check_defaults(config)
                except json.decoder.JSONDecodeError:
                    if self.defaults:
                        return self.defaults
                    else:
                        self.error = f'{text.INVALID_JSON} {self.path}'
        except FileNotFoundError:
            if self.defaults:
                return self.defaults
            else:
                self.error = text.DEFAULTS_ERR
        self.status = self.STATUS_ERROR
        return {}

    def save(self):
        if not self.path.parent.is_dir():
            self.create_directories()
        try:
            with open(self.path, 'w') as f_config:
                json.dump(self.__dict__['config'], f_config)
            return self.OK
        except Exception as err:
            self.status = self.STATUS_ERROR
            self.error = err



"""Utilities for psiconfig."""


def invert(enum: dict) -> dict:
    """Add the inverse items to a dictionary."""
    output = {}
    for key, item in enum.items():
        output[key] = item
        output[item] = key
    return output



"""Base class for config object for the application."""

from pathlib import Path

from .text import FIELD, NOT_IN_DICT


class Config():
    """
    The class takes a path to a config file and if valid returns a config dict.

    Attributes
    ----------

    path: str
        The path to the config file

    defaults: dict[str, object]
        The defaults are used if the path does not contain a valid config file.
    """

    STATUS_OK = 0
    STATUS_ERROR = 1
    STATUS_NOT_DEFINED = -1

    def __init__(
            self,
            path: str,
            defaults: dict[str, str] = {}
            ):
        self.path: str = path
        self.defaults: dict = defaults
        self.status: int | str = self.STATUS_NOT_DEFINED
        self.error: str = ''
        self.config = self._get_config()
        for key, item in self.config.items():
            self.__dict__[key] = item

    def __repr__(self):
        output = ['Config:']
        for key, item in self.__dict__.items():
            output .append(f'{key}: {item}')
        return '\n'.join(output)

    def _get_config(self) -> dict[str, object]:
        """Return config, if contents are valid."""
        config = self._read_config()
        for key, item in config.items():
            self.__dict__[key] = item

        if config:
            return config

        if self.defaults:
            return self.defaults
        return {}

    def update(self, field: str, value: object, force: bool = False) -> None:
        """Update the value of an attribute in config."""
        if not force and field not in self.__dict__['config']:
            self.status = self.STATUS_ERROR
            self.error = f'{FIELD} {field} {NOT_IN_DICT}'
            return

        self.__dict__['config'][field] = value

    def create_directories(self) -> bool:
        """Create directories recursively."""
        create_parts = []
        create_path = Path(self.path).parent
        for part in create_path.parts:
            create_parts.append(part)
            new_path = Path(*create_parts)
            if not Path(new_path).is_dir():
                try:
                    Path(new_path).mkdir()
                except FileExistsError:
                    continue
                except PermissionError:
                    self.status = self.STATUS_ERROR
                    self.error = f'Invalid file path: {new_path}'
                    return False
        return True

    def check_defaults(self, config: dict) -> dict:
        """Make sure all default items in config."""
        for key, item in self.defaults.items():
            if key not in config:
                config[key] = item
        self.config = config
        return config

    @property
    def NOT_DEFINED(self):
        return self.STATUS_NOT_DEFINED

    @property
    def OK(self):
        return self.STATUS_OK

    @property
    def ERROR(self):
        return self.STATUS_ERROR



"""A module to replace tomli and tomli_w for simple cases."""
import re
from icecream import ic


valid_keys_re = [
    r'^\w{1,}$',
    r'^"\w{1,}"$',
    r"^'\w{1,}'$",
]

valid_table_re = [
    r'^\[\w{1,}\]$',
    r'^\["\w{1,}"]$',
    r"^\['\w{1,}']$",
]

valid_equals_re = [
    r'^\w{1,}=\w{1,}$',
    r'^\w{1,}\s{0,}=\s{0,}\w{0,}\s{0,}"\w{0,}\s{0,}=\w{0,}"',
    r"^\w{1,}\s{0,}=\s{0,}\w{0,}\s{0,}'\w{0,}\s{0,}=\w{0,}'",
]

quoted_hash_re = [
    r"\w{0,}'\w{0,}#\w{0,}'\w{0,}",
    r'\w{0,}"\w{0,}#\w{0,}"\w{0,}',
]

valid_number_re = r'^-{0,1}[0-9]{0,}\.{0,1}[0-9]{0,}$'


class TOMLDecodeError(Exception):
    """Exception raised for custom error in the application."""

    def __init__(self, message: str = '') -> None:
        super().__init__(message)

    def __str__(self):
        return f'TOMLDecodeError: {self.message}'


def load(file_handle) -> list:
    """Read and parse a text file and return a dict."""
    try:
        text = file_handle.read()
        return parse(text.split('\n'))
    except Exception as err:
        raise Exception(err.args[0])


def dump(dict: dict, file_handle) -> None:
    """Write dict in TOML format."""
    try:
        text = _dict_to_list(dict)
        file_handle.write(text)
    except Exception as err:
        raise Exception(err.args[0])


def parse(data: list) -> dict:
    result = {}
    for index, text in enumerate(data):
        try:
            (key, item) = _parse(index, text)
            if key:
                result[key] = item
        except TOMLDecodeError as err:
            raise TOMLDecodeError(err.args[0])
    return result


def _parse(line: int, text: str) -> tuple:
    (key, item) = ('', '')
    if not text:
        return (key, item)

    # Comments
    if text[0] == '#':
        return (key, item)
    if '#' in text:
        for test in quoted_hash_re:
            if re.search(test, text):
                break
        else:  # after or no break
            text = text[:text.index('#')]

    # Tables
    if '[' in text or ']' in text or '=' not in text:
        _validate_tables(line, text)

    # Standard items
    if '=' in text:
        if text.count('=') > 1:
            _validate_equals(line, text)
        index = text.index('=')
        key = _get_key(line, text[:index])
        item = _get_item(line, text[index+1:].strip())
    return (key, item)


def _get_key(line: int, key: str) -> str:
    key = key.strip()
    for test in valid_keys_re:
        if re.search(test, key):
            break
    else:  # after or no break
        raise TOMLDecodeError(f'Invalid key definition: line {line}')
    key = key.replace('"', '')
    key = key.replace("'", '')
    return key


def _get_item(line: int, item: str) -> str:
    item = item.strip()
    if (item[0] == '"' and item[-1] == '"'
            or item[0] == "'" and item[-1] == "'"):
        item = item[1:-1]
        return item
    if re.search(valid_number_re, item):
        if '.' in item:
            return float(item)
        return int(item)
    return item


def _validate_equals(line: int, text: str) -> None:
    for test in valid_equals_re:
        if re.search(test, text):
            break
    else:  # after or no break
        raise TOMLDecodeError(f'Invalid equals definition: line {line}')


def _validate_tables(line: int, text: str) -> None:
    for test in valid_table_re:
        if re.search(test, text):
            break
    else:  # after or no break
        raise TOMLDecodeError(f'Invalid table definition: line {line}')


def _dict_to_list(dict) -> str:
    toml = []
    for key, item in dict.items():
        if isinstance(item, str) and re.search(valid_number_re, item):
            item = f'"{item}"'
        toml.append(f'{key} = {item}')
    return '\n'.join(toml)



""" Text for the application."""

# Error messages
LOCATION_ERR = 'The config file is not in the expected location:'
DEFAULTS_ERR = 'No default config defined'
INVALID_TOML = 'Invalid toml format in'
INVALID_JSON = 'Invalid json format in'
CORRUPT_FILE = 'Corrupt config file.'
MISSING_ATTR = 'Missing attribute:'
NOT_OF_TYPE = 'not of type'
FIELD = 'Field'
NOT_IN_DICT = 'not in __dict__'
